import pandas as pd
from scipy import stats
from statsmodels.stats.multicomp import MultiComparison
from statsmodels.stats.anova import AnovaRM
import tabulate as tb
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import json

class Stat_Manager:
    """ args: pandas.DataFrame"""
    
    def __init__(self, dataframe: pd.DataFrame, id: str = None):
        self.df = dataframe
        self.result = None
        if self.df.index.name == None:
            self.df.set_index(id, inplace=True)
        
        else:
            pass
        
        self.id_var = self.df.index.name

        self.menu = {
        'kstest' : {
            'name' : 'Kolmogorov-Smirnov Test',
            'type' : '정규성',
            'group' : 1,
            'testfunc' : stats.kstest,
            'division' : None,
        },
        
        'shapiro' : {
            'name' : 'Shapiro-Wilks Test',
            'type' : '정규성',
            'group' : 1,
            'testfunc' : stats.shapiro,
            'division' : None,
        },
        
        'levene' : {
            'name' : 'Levene Test',
            'type' : '등분산성',
            'group' : 2,
            'testfunc' : stats.levene,
            'division' : None,
        },
        
        'ttest_ind' : {
            'name' : 'Indenpendent Samples T-test',
            'type' : '차이비교_집단간',
            'group' : 2,
            'testfunc' : stats.ttest_ind,
            'division' : '모수'
        },
        
        'ttest_rel' : {
            'name' : 'Dependent Samples T-test',
            'type' : '차이비교_집단내',
            'group' : 1,
            'testfunc' : stats.ttest_rel,
            'division' : '모수'
        },
        
        'mannwhitneyu' : {
            'name' : 'Mann-Whitney U Test',
            'type' : '차이비교_집단간',
            'group' : 2,
            'testfunc' : stats.mannwhitneyu,
            'division' : '비모수'
        },
        
        'wilcoxon' : {
            'name' : 'Wilcoxon-Signed Ranksum Test',
            'type' : '차이비교_집단내',
            'group' : 1,
            'testfunc' : stats.wilcoxon,
            'division' : '비모수'        
        },
        
        
        'f_oneway' : {
            'name' : 'One-way ANOVA',
            'type' : '차이비교_집단간',
            'group' : 3,
            'testfunc' : stats.f_oneway,
            'division' : '모수'
            
        },
        
        'kruskal' : {
            'name' : 'Kruskal-Wallis Test',
            'type' : '차이비교_집단간',
            'group' : 3,
            'testfunc' : stats.kruskal,
            'division' : '비모수'
            
        },
        
        'chi2_contingency' : {
            'name' : 'Chi-Square Test',
            'type' : '빈도분석',
            'group': 1,
            'testfunc' : stats.chi2_contingency,
            'division' : None
            },
        
        'z_normal' : {
            'name' : 'z-skeweness & z-kurtosis test',
            'type' : '정규성_예외',
            'group' : 1,
            'testfunc' : self.zscore_normality,
            'division' : None
            },
        
        'fmax' : {
            'name' : 'F-max Test',
            'type' : '등분산성_예외',
            'group' : 2,
            'testfunc' : self.fmax_test,
            'division' : None,
            },
        
        'pearsonr' : {
            'name' : '상관분석: Pearson r',
            'type' : '상관분석',
            'group' : 1,
            'testfunc' : self.r_forargs,
            'division' : None,
        },
        
        'spearmanr' : {
            'name' : '상관분석: Spearman r',
            'type' : '상관분석',
            'group' : 1,
            'testfunc' : self.r_forargs,
            'division' : None,
        },
        'friedman' : {
            'name' : 'Friedman Test',
            'type' : '차이비교_집단내',
            'group' : 1,
            'testfunc' : stats.friedmanchisquare,
            'division' : '비모수'
        },
        'f_oneway_rm' : {
            'name' : 'One-way Repeated Measures ANOVA',
            'type' : '차이비교_집단내',
            'group' : 1,
            'testfunc' : AnovaRM,
            'division' : '모수'
        },
        'bootstrap1000' : {
            'name' : 'Bootstrap percentile method: Resampling No. = 1,000',
            'type' : '차이비교_예외',
            'group' : 1,
            'testfunc' : self.percentile_method,
            'division' : None
        },
        'bootstrap10000' : {
            'name' : 'Bootstrap percentile method: Resampling No. = 10,000',
            'type' : '차이비교_예외',
            'group' : 1,
            'testfunc' : self.percentile_method,
            'division' : None
        }
    }
        
    def progress(self, method: str, vars: list, group_vars: str = None, group_names: list = None, posthoc: bool = False):
        df = self.df.dropna(axis=0, how = 'any', subset = vars)
        testfunc = self.menu[method]['testfunc']
        group_fill = self.menu[method]['group']
        testtype = self.menu[method]['type']
        testname = self.menu[method]['name']
        testdivision = self.menu[method]['division']
        
        n = len(df)
        
        if testtype == '빈도분석':
            ser = pd.crosstab(df[vars[0]], df[vars[1]])
            re = testfunc(ser)
            s = re[0]
            p = re[1]
            s = round(s, 3)
            p = round(p, 3)

            print(f"\n{testname}")
            print(f"변수 : {vars[0]}, {vars[1]}")
            print(f"\n검정통계치 χ² = {s:.3f}, p = {p:.3f}\n")
            
            self.result = tb.tabulate(ser, tablefmt='pipe', headers=ser.columns.to_list())
            print(self.result)
            
            # return ser
        
        if testtype == '정규성':
            ser = df[vars[0]]
            print(f"\n{testname}")
            if method == 'kstest':
                s, p = testfunc(ser, 'norm')
                
                if n < 30:
                    print("주의: 표본 수가 30보다 적습니다. 다른 분석을 고려하십시오. ")
                
            else:
                s, p = testfunc(ser)
                
                if n >= 30:
                    print("주의: 표본 수가 30보다 많습니다. 다른 분석을 고려하십시오.")
            
            
            s = round(s, 3)
            p = round(p, 3)
            
            print(f"변수 : {vars[0]}")
            print(f"n = {n}")
            print(f"검정통계치 = {s:.3f}, p = {p:.3f}")
            
            if p <= .05 : 
                print("정규성 가정 미충족")
                
            else:
                print('정규성 가정 충족')
        
        if testtype == '등분산성':
            series = []
            for n in range(len(group_names)):
                ser = df.loc[df[group_vars] == group_names[n], vars[0]]
                series.append(ser)
            
            s, p = testfunc(*series)
            s = round(s, 3)
            p = round(p, 3)
            
            print(f"\n{testname}")
            print(f"집단변수 : {group_vars}")
            print(f"비교집단 : {group_names}")
            print(f"\n검정통계치 = {s:.3f}, p = {p:.3f}")
            
            
            if p <= .05 :
                print('등분산성 가정 미충족')
            else:
                print('등분산성 가정 충족')
                
        if testtype == '차이비교_집단내':
            
            if method == 'friedman' or method == 'f_oneway_rm':
                series = []
                for n in range(len(vars)):
                    ser = df[vars[n]]
                    series.append(ser)

                dict = {}
                for n in range(len(vars)):
                    dict[vars[n]] = {
                        'mean' : "{:.2f}".format(series[n].mean().round(2)),
                        'median' : "{:.2f}".format(series[n].median().round(2)),
                        'sd' : "{:.2f}".format(series[n].std().round(2)),                        
                    }
                dict = pd.DataFrame(dict)
                self.result = tb.tabulate(dict, tablefmt ='pipe', headers=dict.columns.to_list())
                
                print(f"\n{testname}")
                print(f"변수 : {vars}, 시점 = {len(vars)}")
                print(f"변수별 기술통계치: \n\n")
                print(self.result)
                
                if method == 'friedman':
                    s, p = testfunc(*series)
                    s = round(s, 3)
                    p = round(p, 3)
                    
                    print(f"\n검정통계치 = {s:.3f}, p = {p:.3f}")
                
                elif method == 'f_oneway_rm':
                    reset_df = df.reset_index().melt(id_vars= self.id_var, value_vars = vars)
                    result = AnovaRM(data = reset_df, depvar = 'value', subject = self.id_var, within = ['variable']).fit()
                    s = result.anova_table['F Value']
                    p = result.anova_table['Pr > F']
                    s = round(s, 3)
                    p = round(p, 3)

                    for f, s, p in zip(result.anova_table.index, s, p):
                        print(f"\n검정통계치 F = {s:.3f}, p = {p:.3f}")

                if posthoc == True:
                    posthoc_df = df.reset_index().melt(id_vars=self.id_var, value_vars=vars)
                    mc = MultiComparison(posthoc_df['value'], posthoc_df['variable'])
                    
                    if testdivision == '모수':
                        result = mc.allpairtest(stats.ttest_ind, method = 'bonf')
                        return result[0]
                    
                    else: 
                        result = mc.allpairtest(stats.mannwhitneyu, method = 'bonf')
                        return result[0]

            
                
            else:
                series = []
                for n in range(len(vars)):
                    ser = df[vars[n]]
                    series.append(ser)
                
                dict = {}
                for n in range(len(vars)):
                    dict[vars[n]] = {
                        'mean' : series[n].mean().round(2),
                        'median' : series[n].median().round(2),
                        'sd' : series[n].std().round(2),                        
                    }
                
                dict = pd.DataFrame(dict)
                self.result = tb.tabulate(dict, tablefmt ='pipe', headers=dict.columns.to_list())
                
                s, p = testfunc(series[0], series[1])
                s = round(s, 3)
                p = round(p, 3)
                
                n = len(df)
                
                print(f"\n{testname}")
                print(f"변수 : {vars[0]}, {vars[1]}")
                print(f"n = {n}\n")
                print(self.result)
                print(f"\n검정통계치 = {s:.3f}, p = {p:.3f}")
            
        if testtype == '차이비교_집단간':
            series = []
            for n in range(len(group_names)):
                ser = df.loc[df[group_vars] == group_names[n], vars[0]]
                series.append(ser)
            
            dict = {}
            for n in range(len(group_names)):
                dict[group_names[n]] = {
                    'n' : len(series[n]),
                    'mean' : series[n].mean().round(2),
                    'median' : series[n].median().round(2),
                    'sd' : series[n].std().round(2),
                    }
            
            dict = pd.DataFrame(dict)
            self.result = tb.tabulate(dict, tablefmt ='pipe', headers=dict.columns.to_list())
            
            s, p = testfunc(*series)
            s = round(s, 3)
            p = round(p, 3)
        
            print(f"\n{testname}")
            print(f"변수 : {vars[0]}")
            print(f"집단변수 : {group_vars}")
            print(f"비교집단 : {group_names}")
            print(f"집단별 기술통계치: \n\n{self.result}\n")
            print(f"검정통계치 = {s:.3f}, p = {p:.3f}")
            
            if posthoc == True:
                
                cond_list = []
                for n in range(len(group_names)):
                    cond = df[group_vars] == group_names[n]
                    cond_list.append(cond)
                
                
                selected_rows = pd.concat(cond_list, axis=1).any(axis=1)
                selected_df = df[selected_rows]
                
                mc = MultiComparison(selected_df[vars[0]], selected_df[group_vars])
                
                
                if testdivision == '모수':
                    result = mc.allpairtest(stats.ttest_ind, method = 'bonf')
                else: 
                    result = mc.allpairtest(stats.mannwhitneyu, method = 'bonf')
                
                print('\nPost-hoc 실시')
                return result[0]
                
                
        if testtype == '차이비교_예외':
            
            if method == 'bootstrap1000':
                resampling_no = 1000
                
            elif method == 'bootstrap10000':
                resampling_no = 10000
            
            if group_vars == None:
            
                print(f"{testname}: \n")
                
                ser1 = self.bootstrap(series = df[vars[0]], n_bootstrap=resampling_no)
                ser2 = self.bootstrap(series = df[vars[1]], n_bootstrap=resampling_no)
                bootstrap_df = self.bootstrap_to_dataframe(ser1, ser2, label = vars)
                testfunc(data = bootstrap_df, a_var = vars[0], b_var = vars[1])
                
            else:
                print(f"{testname}: \n")
                
                ser1 = self.bootstrap(series= df.loc[df[group_vars] == group_names[0], vars[0]], n_bootstrap=resampling_no)
                ser2 = self.bootstrap(series= df.loc[df[group_vars] == group_names[1], vars[0]], n_bootstrap=resampling_no)
                a_var = f"{group_names[0]}_{vars[0]}"
                b_var = f"{group_names[1]}_{vars[0]}"
                bootstrap_df = self.bootstrap_to_dataframe(ser1, ser2, label = [a_var, b_var])
                testfunc(data = bootstrap_df, a_var = a_var, b_var = b_var)
                
                
        if testtype == '정규성_예외':
            print(f"{testname}")
            print(f"변수: {vars[0]}")
            testfunc(self.df[vars[0]])
            
        if testtype == '등분산성_예외':
            print(f"{testname}")
            print(f"변수: {vars[0]}")
            testfunc(vars = vars, group_vars = group_vars, group_names = group_names)
            
        if testtype == '상관분석':
            print(testname)
            print(f"변수 : {vars}\n")
            testfunc(method = method, vars = vars)
            
            if method == 'pearsonr':
                return df[vars].corr().round(3)
    
    def zscore_normality(self, series):
        n = series.count()
        
        skewness = series.skew().round(3)
        skewness_se = np.sqrt(6 * n * (n - 1) / ((n - 2) * (n + 1) * (n + 3))).round(3)
        
        kurtosis = series.kurtosis().round(3)
        kurtosis_se = (np.sqrt((n**2 - 1) / ((n-3)*(n+5))) * skewness_se * 2).round(3)
        
        z_skewness = (skewness/skewness_se).round(3)
        z_kurtosis = (kurtosis/kurtosis_se).round(3)
        
        if n < 50:
            cutoff = 1.96
        elif n < 200:
            cutoff = 2.59
        elif n > 200:
            cutoff = 3.13
        
        print(f"skewness = {skewness}\nstandard error of skewness = {skewness_se}\nz-skewness = {z_skewness}\n\nkurtosis = {kurtosis}\nstandard error of kurtosis = {kurtosis_se}\nz-kurtosis = {z_kurtosis}\n\nsample n = {n}, corresponding absolute cutoff score of z-skewenss and z-kurtosis = {cutoff}")
        
        z_skewness = abs(z_skewness)
        z_kurtosis = abs(z_kurtosis)
        
        print("\n결과: ")
        if z_skewness < cutoff and z_kurtosis < cutoff:
            print("\n정규성 가정 충족")
        else:
            print("\n정규성 가정 미충족")
            
            
        print("\nReferences:\n[1] Ghasemi, A., & Zahediasl, S. (2012). Normality tests for statistical analysis: a guide for non-statisticians. International journal of endocrinology and metabolism, 10(2), 486. \n[2] Moon, S. (2019). Statistics for the Social Sciences: Moving Toward an Integrated Approach. Cognella Academic Publishing.")
        
    def fmax_test(self, vars, group_vars, group_names):
        df = self.df.loc[self.df[group_vars].isin(group_names)]
        group_n = len(group_names)
        
        max_variance = df.groupby(group_vars)[vars[0]].var().max().round(3)
        min_variance = df.groupby(group_vars)[vars[0]].var().min().round(3)
        
        f_max = max_variance / min_variance
        f_max = round(f_max, 3)
        
        
        print(f"\n집단 수 = {group_n}")
        print(f"Max variance among group = {max_variance}")
        print(f"Min variance among group = {min_variance}")
        print(f"F-max statistics = {f_max}\n")
        print("\n결론:")
        
        if f_max < 10:
            print("등분산성 가정 충족")
        else:
            print("등분산성 가정 미충족")
            
        print("\nReference:\n[1] Fidell, L. S., & Tabachnick, B. G. (2003). Preparatory data analysis. Handbook of psychology: Research methods in psychology, 2, 115-141.\n")
    
    def r_forargs(self, method, vars):
        df = self.df.dropna(axis=0, how = 'any', subset = vars)
        
        if method == 'pearsonr':
            tf = stats.pearsonr
        
        elif method == 'spearmanr':
            tf = stats.spearmanr
        
        num = len(vars)
        sets = []
        
        for i in range(num -1):
            for j in range(i +1, num):
                sets.append((df[vars[i]], df[vars[j]]))
            
        for n in sets:
            s, p = tf(n[0], n[1])
            s = round(s, 3)
            p = round(p, 3)
            var1 = n[0].name
            var2 = n[1].name
            
            print(f"{var1} & {var2} r = {s}, p = {p}")
    def bootstrap(self, series, n_bootstrap=1000, statistic=np.mean):
        
        n = len(series)
        bootstrap_results = []
        for _ in range(n_bootstrap):
            bootstrap_sample = series.sample(n, replace=True)  # 재표집
            statistic_value = statistic(bootstrap_sample)
            bootstrap_results.append(statistic_value)
        
        return bootstrap_results
    
    def bootstrap_to_dataframe(self, *args, label):
        n = len(args)
        dict = {}
        for _ in range(n):
            key = f"{_}"
            value = args[_]
            dict[key] = value
        result = pd.DataFrame(dict)
        
        if label != None and type(label) == list:
            t = len(label)
            for n in range(t):
                result.rename(columns = {f'{n}' : label[n]}, inplace=True)
                
        return result

    def percentile_method(self, data, a_var, b_var, confidence_level = 0.95, hist=True):
        
        confidence_dict = {
            0.90 : [5, 95],
            0.95 : [2.5, 97.5],
            0.99 : [0.5, 99.5],
        }
    
        interval = confidence_dict[confidence_level]
        
        n = len(data)
        
        a_confidence_interval = np.percentile(data[a_var], interval)
        b_confidence_interval = np.percentile(data[b_var], interval)
        a_lower_bound = a_confidence_interval[0]
        a_upper_bound = a_confidence_interval[1]
        b_lower_bound = b_confidence_interval[0]
        b_upper_bound = b_confidence_interval[1]
        
        
        
        print(f"{a_var} 의 {confidence_level * 100:.0f}% 신뢰구간: [{a_lower_bound:.3f}, {a_upper_bound:.3f}]")
        print(f"{b_var} 의 {confidence_level * 100:.0f}% 신뢰구간: [{b_lower_bound:.3f}, {b_upper_bound:.3f}]\n")
        
        
        if a_upper_bound < b_lower_bound or a_lower_bound > b_upper_bound:
            print("두 분포의 신뢰구간이 중복되지 않습니다. \n두 분포 간 차이가 유의합니다.")
        else:
            print("두 분포의 신뢰구간이 중복됩니다. \n두 분포 간 차이가 유의하지 않습니다.")
        
        print("\nReference:\nEfron, B., & Tibshirani, R. (1986). Bootstrap methods for standard errors, confidence intervals, and other measures of statistical accuracy. Statistical Science, 1(1), 54-75.\n")
        print("히스토그램: \n")
        
        
        if hist == True:
            plt.figure(figsize=(10, 8))
            sns.set(font = "Gulim", font_scale = 1.5)
            plt.style.use('grayscale')
            plt.title(f'Histogram of {a_var} & {b_var}')
            sns.histplot(data = data[a_var], label = a_var, alpha=0.5, kde=True)
            sns.histplot(data = data[b_var], label = b_var, alpha=0.5, kde=True)
            
            plt.axvline(a_lower_bound, color='black', linestyle='--', label=f'{confidence_level * 100:.0f}% CI ({a_var})')
            plt.axvline(a_upper_bound, color='black', linestyle='--')
            plt.axvline(b_lower_bound, color='gray', linestyle='--', label=f'{confidence_level * 100:.0f}% CI ({b_var})')
            plt.axvline(b_upper_bound, color='gray', linestyle='--')
                    
            plt.xlabel(f"Value of {a_var} & {b_var}")
            plt.ylabel("No. of Samples")
            plt.legend(bbox_to_anchor=(1, 1))
            plt.grid(False)
            plt.show()
            
    def howtouse(self):
        print('something')